function [CorrelationMatrix,CorrelationMatrixP,PvaluesMatrix]=CorrDrag(Wavenumber,DataMatrix,pvaluelim)
% This function creates a figure that shows the correlation of any
% wavenumber (chosen by clicking on the relevant point on the data average
% in the figure below) with the remaining wavenumbers on the data.
%
%% INPUTS
%   Wavenumber: vector with the wavenumbers for each of the Raman
%   intensities
%   DataMatrix: Matrix with spectra as columns, that is, rows as
%   wavenumbers and columns as observations
%   pvaluelim: limit for the p-value - any correlation with a p-value
%   higher than this will be set to zero. Default value is 0.0001 if unset.
%
%% OUTPUTS
%   CorrelatioMatrix: see help for corrcoef
%   CorrelationMatrixP: filtered correlation matrix, where values with a
%   Pvalue higher than 0.0001 are set to 0.
%   P values matrix: see help for corrcoef
%   Draggable figure

if nargin<3
  pvaluelim = 0.0001;
  disp('No pvalue limit was set, default p<0.0001 is used.')
end

try
    disp('Calculating the correlation matrix')
    [CorrelationMatrix PvalueMatrix]=corrcoef(DataMatrix.');
catch
    error('Error with calculating the correlation matrix')
end

try
    disp('Calculating the p-filtered correlation matrix')
    CorrelationMatrixP=CorrelationMatrix;
    for i=1:size(PvalueMatrix,1)
        for j=1:size(PvalueMatrix,1)
            if i>j
                if PvalueMatrix(i,j)>=pvaluelim
                    CorrelationMatrixP(i,j)=0;
                    CorrelationMatrixP(j,i)=0;
                end
            end
        end
    end
catch
    error('Error with p-value filtering the correlation matrix')
end

% Opening draggable figure
try
    figure
    while 1==1
        subplot(2,1,1)
        hold off
        AverageAll=mean(DataMatrix,2);
        plot(Wavenumber,AverageAll,'k')
        hold on
        xlim([Wavenumber(1) Wavenumber(end)])
        [j,~] = ginput(1);
        plot(j,AverageAll(find(round(j)==Wavenumber,1)),'ok')

        subplot(2,1,2)
        indextemp=find(round(j)==Wavenumber,1);
        plot(Wavenumber,CorrelationMatrixP(:,indextemp));
        hold on
        line([Wavenumber(1) Wavenumber(end)],[0.3 0.3],'Color','k')
        line([Wavenumber(1) Wavenumber(end)],[-0.3 -0.3],'Color','r')
        xlabel('Wavenumber (cm^{-1})')
        ylabel('Correlation')

        % Find peaks and label them
        [Peak, PeakIdx] = findpeaks(CorrelationMatrixP(:,indextemp),'MinPeakProminence',0.2,'MinPeakHeight',0.25); 
        for i=1:length(Peak)
            text(Wavenumber(PeakIdx(i)), Peak(i), sprintf('%6.1f', Wavenumber(PeakIdx(i))));
        end    
        plot(Wavenumber(PeakIdx), Peak,'ok')

        legend(num2str(j))
        xlim([Wavenumber(1) Wavenumber(end)])
        hold off
    end
catch
    error('Error with figure, did you delete it?')
end