%% SCRIPT FOR DATASET CREATION FOR RAMAN SINGLE-CELL DATA
% Julia Gala de Pablo
% PhD Student
% University of Leeds

% WDF FILE IMPORTING:
% Cosmic rays should be corrected from the wdf files before using this
% script.

clear all
close all
clc

% Add one dirPath per dataset:
dirPath={...   
    'C:\DirPathExample';...
    };

% Saving Path is common for all datasets:
savingPath='C:\SavingPathExample';

% For baseline subtraction, add the path where the function
% f_baseline_corr is (Koch et al, 2017) and the WiRe import functions are:
addpath('C:\MatlabFunctionsPath')

% The file names are always an identifier + a number, that starts counting
% at 0. Data should be cosmic ray corrected in the WiRe software and
% exported as txt. If acquired independently, the "Create maps from
% singles" tool can be used for quicker joining of wdf. files and
% exporting.

% Cells day 1
Name{1,1}='Celltype_group_';
Index{1}=[1:100]-1;
% Example of the actual peak position for that experiment:
CalibrationPeak{1}=520.482;
NameSaving{1}='Experiment1_';
BName{1}='Background_';
BIndex{1}=1:6;

% Datasets to run:
selectedDataset=[1];

% Creating the list of Filenames and double checking that the files exist 
% and the indexes are correct:
for i=1:length(Name)
    cd(dirPath{i})
    FileNames{i}=cellstr(ls(strcat(Name{i},'*')));
    BNames{i}=cellstr(ls(strcat(BName{i},'*')));
    if length(FileNames{i})>length(Index{i})
        disp(strcat('More files than expected for dataset - ',num2str(i)))
    elseif length(FileNames{i})<length(Index{i})
        disp(strcat('Less files than expected for dataset - ',num2str(i)))
    end
    if length(BNames{i})>length(BIndex{i})
        disp(strcat('More files than expected for dataset backgrounds - ',num2str(i)))
    elseif length(BNames{i})<length(BIndex{i})
        disp(strcat('Less files than expected for dataset backgrounds - ',num2str(i)))
    end    
end

fprintf('Loading the Data \n')
tic

%% Loading the data
fprintf('Loading the Data \n')
tic
for j=selectedDataset
    try
        FileName={};
        m=1;
        b=1;
        fprintf(['Loading data from dataset' num2str(j) '\n']);
        cd(dirPath{j});
        % Importing the data for the spectra
        for i=1:length(FileNames{j})
            % Importing the wdf
            wdf=WdfReader(FileNames{j}{i});
            RamanSpectra{m}(:,1)=wdf.GetXList;
            RamanSpectra{m}(:,2)=wdf.GetSpectra(1,wdf.Count);
            m=m+1;
            clear wdf
        end
        % Importing the background spectrum
        for i=1:length(BNames{j})
            wdf=WdfReader(BNames{j}{i});
            RamanBackgroundSpectra{i}(:,1)=wdf.GetXList;
            RamanBackgroundSpectra{i}(:,2)=wdf.GetSpectra(1,wdf.Count);
            clear wdf
        end

        CalibrationPeakTemp=CalibrationPeak{j};
        
        % Saving the raw data in an easy/quick to load .mat file.
        cd(savingPath)
        save([NameSaving{j} '_datasetraw'],...
            'RamanSpectra','CalibrationPeakTemp','RamanBackgroundSpectra')
        toc

        %% X alignment of the spectra respect to the Calibration peak position in Si

        % Every day of measuring a calibrating spectra is taken and it is aligned
        % as good as it can to the 520.5 cm^-1 peak of Silicon. However, sometimes
        % the peak seems to be around 520.3 or 520.8. This peak will be used to
        % align the Shift of the spectra and reduce the variability between days.
        % As the PCA only considers the Y-values for analysis, a spline function 
        % should be fit so the Y-values have exactly the same wavenumbers.
        fprintf('Aligning spectra and fitting to the selected X values using a spline \n')
        tic
        Xintvalues=300:0.5:3200; %X values for the interpolation
        RamanSpectraInt=cell(size(RamanSpectra));

        for k=1:length(RamanSpectra)
           RamanSpectraInt{k}=spline(RamanSpectra{k}(:,1)+...
               (520.5-CalibrationPeak{j}),RamanSpectra{k}(:,2),...
               Xintvalues);
        end
        for k=1:length(RamanBackgroundSpectra)
           RamanBSpectraInt{k}=spline(RamanBackgroundSpectra{k}(:,1)+...
               (520.5-CalibrationPeak{j}),RamanBackgroundSpectra{k}(:,2),...
               Xintvalues);
        end
        toc

        %% Background Correction with the backgrounds obtained

        disp('Subtracting the quartz background')

        % Calculate the average of the backgrounds available
        BackgroundAverage=mean(cell2mat(RamanBSpectraInt.'));

        % Subtract the baseline of the Background
        BackgroundAverageZero=f_baseline_corr(Xintvalues,...
                BackgroundAverage,40,400,5);

        % Background subtraction: the best way would be to fit the spectra to the
        % reference background, obtaining an index, and then subtracting using that
        % index with a sum of squares method.

        % Function
        bb=spline(Xintvalues,BackgroundAverageZero);
        f=@(a,x) a*ppval(bb,x);

        % Adding weights focuses the fit onto the areas with stronger background contribution
        f0 = fitoptions('Weights',BackgroundAverageZero.^6,'Exclude',Xintvalues>2000);

        % Subtract the average background
        for k=1:length(RamanSpectra)
            % Baseline subtration
            RamanSpectraTemp=f_baseline_corr(Xintvalues,...
                RamanSpectraInt{k},40,400,5);
            IntPeak=fit(Xintvalues.',RamanSpectraTemp,f,f0);
            RamanSpectraBackgSubs{k}=RamanSpectraTemp.'-...
                (BackgroundAverageZero.')*IntPeak.a;
            clear IntPeak
        end

        %% Smoothing of the spectra
        fprintf('Smoothing the spectra \n')
        tic
        RamanSpectraSmooth=cell(size(RamanSpectraBackgSubs));
        for k=1:size(RamanSpectraBackgSubs,2)
            RamanSpectraSmooth{k}=sgolayfilt(RamanSpectraBackgSubs{k},2,17);
        end
        toc

        %% Truncating the spectra
        % Removing non-desired parts of the spectrum
        fprintf('Truncating the spectra \n')
        tic
        lowerlim=500;
        higherlim=3080;
        lowerlimIndex=find(floor(Xintvalues)==lowerlim,1);
        higherlimIndex=find(floor(Xintvalues)==higherlim,1);
        XvaluesTrunc=[Xintvalues(lowerlimIndex:higherlimIndex)];
        
        for k=1:length(RamanSpectraSmooth)
            RamanSpectraTrunc{k}=RamanSpectraSmooth{k}(lowerlimIndex:higherlimIndex);
        end
        toc

        %% Subtract the baseline
        fprintf('Subtracting the baseline \n')
        % Here we use the function f_baseline_corr from Koch et al., 2017.
        tic
        RamanSpectraBaseline=cell(size(RamanSpectraInt));
        for i=1:size(RamanSpectra,2)   
            disp(['Correcting spectrum ' num2str(i) ' of ' num2str(length(RamanSpectraBaseline))])
            % First step to remove the maximum possible of the baseline
            [RamanSpectraB{i}, varargout{i}]=f_baseline_corr(XvaluesTrunc,...
                RamanSpectraTrunc{i},40,600,10);
            % Second step to help remove the constant baseline 
            [RamanSpectraB2{i}, varargout{i}]=f_baseline_corr(XvaluesTrunc,...
                RamanSpectraB{i},60,800,3);
            RamanSpectraBaseline{i}=RamanSpectraB2{i}.';
        end
        toc

        %% Elimination of regions with no interest and high background

        fprintf('Zapping the spectra \n')
        tic
        zappingIndex=find((1750<XvaluesTrunc)&(XvaluesTrunc<2130));
        zappingIndex2=find((2285<XvaluesTrunc)&(XvaluesTrunc<2750));
        XvaluesTrunc2=[XvaluesTrunc(1:min(zappingIndex(:)))...
                    XvaluesTrunc(max(zappingIndex(:)):min(zappingIndex2(:)))...
                    XvaluesTrunc(max(zappingIndex2(:)):end)];

        for i=1:length(RamanSpectraBaseline)
                RamanSpectraZap{i}=[RamanSpectraTrunc{i}(1:min(zappingIndex(:))) ...
                    RamanSpectraTrunc{i}(max(zappingIndex(:)):min(zappingIndex2(:))) ...
                    RamanSpectraTrunc{i}(max(zappingIndex2(:)):end)];   
        end
        toc

        %% Normalization of the spectra
        fprintf('Normalizing the spectra \n')
        tic
        % Intensity of the peaks may vary due to different setups, so only relative
        % intensity between peaks should be considered.

        RamanSpectraNorm={};

        for i=1:length(RamanSpectraZap)
          % For normalization to the amide peak
          x1=find(round(XvaluesTrunc2)==1640);
          x2=find(round(XvaluesTrunc2)==1680);
          % We want to normalize to a very smooth version of the spectrum.
          RamanSpectraNorm{i}=RamanSpectraZap{i}/...
              max(sgolayfilt(RamanSpectraZap{i}(x1(1):x2(1)),2,31));       
        end
        toc

        %% Data Matrix
        % Get the colums of intensities to analyse in a matrix:
        fprintf('Creating the Data Matrix \n')
        tic
        DataMatrix=[];
        for i=1:size(RamanSpectraNorm,2)
            DataMatrix(:,end+1)=RamanSpectraNorm{i};
        end
        toc

        %% Calculate the Average and STD
        Average=mean(DataMatrix,2);
        Error=std(DataMatrix,0,2);

        %% Saving Data
        disp('saving data')
        cd(savingPath)
        save([NameSaving{j} '_datasetanalysed'],...
            'DataMatrix','Average','Error','XvaluesTrunc2')

        % Clear data for next dataset
            clear RamanSpectra RamanSpectraB RamanBackgroundSpectra ...
                RamanBSpectraInt RamanSpectraBaseline RamanSpectraNorm ...
                RamanSpectraQuartz RamanSpectraSmooth RamanSpectraTrunc ...
                RamanBSpectraInt RamanSpectraInt RamanSpectraZero...
                CalibrationPeakTemp BackgroundAverage BackgroundAverageZero...
                DataMatrix RamanSpectraZap
    catch
        disp(['Something went wrong for Dataset: ',num2str(j),...
            '// proceeding with next dataset'])
    end        
end

disp('Done :)')

